# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D


RPMS_OF_INTEREST_RE = re.compile(
    r"^(\S*(?:openvswitch|ovn|dpdk|tuned|libvirt|qemu)\S*)\s", re.MULTILINE
)
PODMAN_PS_RE = re.compile(r"^[a-f0-9]+\s+(\S+)\s", re.MULTILINE)


def parse_report(path: pathlib.Path, data: D):
    data.software = sw = D()
    data.hostname = (path / "hostname").read_text().strip()

    for f in (
        "etc/os-release",
        "etc/rhosp-release",
    ):
        f = path / f
        if f.is_file():
            value = f.read_text().strip()
            if f.name == "os-release":
                match = re.search(r"^PRETTY_NAME=(.+)$", value, flags=re.MULTILINE)
                if match:
                    value = match.group(1).strip("'\" ")
            sw[f.name.replace("-", "_")] = value
    rpms = set()
    for match in RPMS_OF_INTEREST_RE.finditer((path / "installed-rpms").read_text()):
        rpms.add(match.group(1))
    sw.rpms = rpms

    containers = set()
    podman_ps = path / "sos_commands/podman/podman_ps"
    if podman_ps.is_file():
        for match in PODMAN_PS_RE.finditer(podman_ps.read_text()):
            _, _, name = match.group(1).partition("/")
            if "libvirt" in name or "qemu" in name or "ovn" in name:
                containers.add(name)
    sw.containers = containers

    match = re.search(r"/vmlinuz-(\S+)", (path / "proc/cmdline").read_text())
    if match:
        sw.kernel = match.group(1)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D
from ..bits import parse_cpu_set


PORT_RE = re.compile(
    r"""
    ^\s{8}Port\s(?P<name>.+)\n
        (?:^\s+tag:\s(?P<tag>\d+)\n)?
        (?P<ifaces>(?:^\s+Interface\s.+\n
            (?:^\s+type:\s.*\n)?
            (?:^\s+options:\s.*\n)?
        )+)
    """,
    re.VERBOSE | re.MULTILINE,
)
IFACE_RE = re.compile(
    r"""
    ^\s+Interface\s+(?P<name>.+)\n
        (?:^\s+type:\s+(?P<type>.+)\n)?
        (?:^\s+options:\s+(?P<options>.+)\n)?
    """,
    re.VERBOSE | re.MULTILINE,
)
RXQ_RE = re.compile(
    r"""
    \s+port:\s+(\S+)
    \s+queue-id:\s+(\d+)
    (?:\s+\((?P<status>enabled|disabled)\))?
    \s+pmd\susage:\s*(\d+)\s*%
    $
    """,
    re.VERBOSE | re.MULTILINE,
)


def parse_report(path: pathlib.Path, data: dict):
    data.ovs = ovs = D()
    ovs.config = conf = D()
    for f in path.glob("sos_commands/openvswitch/ovs-vsctl*_list_Open_vSwitch"):
        ovs_to_dict(f.read_text(), conf)
    if "other_config" in conf and "pmd_cpu_mask" in conf.other_config:
        mask = conf.other_config.pmd_cpu_mask
        if not mask.startswith("0x"):
            mask = "0x" + mask
        conf.dpdk_cores = parse_cpu_set(mask)

    ovs_ports(ovs, path)
    ovs_pmds(ovs, path)


def ovs_ports(ovs, path):
    ovs.bridges = bridges = D()
    ovs.ports = ports = D()
    for f in path.glob("sos_commands/openvswitch/ovs-vsctl*_show"):
        for block in re.split(r"^    Bridge ", f.read_text(), flags=re.MULTILINE):
            br_name, block = block.split("\n", 1)
            br_name = strip_quotes(br_name)
            datapath = "system"
            match = re.search(r"datapath_type: (\S+)", block)
            if match:
                datapath = match.group(1)
            for match in PORT_RE.finditer(block):
                ifaces = D()
                for m in IFACE_RE.finditer(match.group("ifaces")):
                    name = strip_quotes(m.group("name"))
                    ifaces[name] = D(name=name, type=m.group("type") or "")
                    if m.group("options"):
                        ifaces[name].options = cast_value(m.group("options"))
                if not ifaces:
                    continue
                port_name = strip_quotes(match.group("name"))
                port = D(name=port_name, bridge=br_name)
                if len(ifaces) == 1 and port_name in ifaces:
                    port.update(ifaces[port_name])
                else:
                    port.type = "bond"
                    port.members = ifaces
                if match.group("tag"):
                    port.tag = int(match.group("tag"))

                ports[port_name] = port
                bridges.setdefault(
                    br_name, D(name=br_name, ports=0, of_rules=0)
                ).ports += 1
                bridges[br_name].datapath = datapath

    for name, br in bridges.items():
        for f in path.glob(f"sos_commands/openvswitch/ovs-ofctl*_dump-flows_{name}"):
            br.of_rules = len(f.read_text().splitlines()) - 1


def strip_quotes(s: str) -> str:
    return s.strip("\"' \t")


def ovs_pmds(ovs, path):
    ovs.pmds = pmds = D()
    f = path / "sos_commands/openvswitch/ovs-appctl_dpif-netdev.pmd-rxq-show"
    if f.is_file():
        for block in re.split(r"^pmd ", f.read_text(), flags=re.MULTILINE):
            match = re.search(r"thread numa_id (\d+) core_id (\d+):", block)
            if not match:
                continue
            numa, core = match.groups()
            core = int(core)
            numa = int(numa)
            pmds[core] = D(numa=numa, core=core, rxqs=[])

            match = re.search(r"isolated\s*:\s*(true|false)", block)
            if match:
                pmds[core].isolated = match.group(1) == "true"

            for match in RXQ_RE.finditer(block):
                port, rxq, status, usage = match.groups()
                pmds[core].rxqs.append(
                    D(
                        port=strip_quotes(port),
                        rxq=int(rxq),
                        usage=int(usage),
                        enabled=status in ("enabled", None, ""),
                    )
                )


PROP_RE = re.compile(r"^([\w-]+)\s*:\s*(.*)$", re.MULTILINE)


def ovs_to_dict(block: str, d: dict):
    for match in PROP_RE.finditer(block):
        key, value = match.groups()
        value = cast_value(value)
        if value in ("", [], {}):
            continue
        d[strip_quotes(key.replace("-", "_"))] = value


def cast_value(value: str):
    value = value.strip("\t\n\r ")
    if value.startswith("["):
        data = []
        for token in value.strip("[]").split(", "):
            if token.strip() != "":
                data.append(cast_value(token))
        return data
    if value.startswith("{"):
        data = D()
        for token in value.strip("{}").split(", "):
            if token.strip() != "":
                try:
                    key, val = token.split("=", 1)
                except ValueError:
                    continue
                data[key.strip().replace("-", "_")] = cast_value(val)
        return data
    if value.strip('"') == "true":
        return True
    if value.strip('"') == "false":
        return False
    if value.isdigit():
        return int(value)
    return strip_quotes(value)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

from . import dot, json, svg, text


FORMATS = {"dot": dot, "text": text, "json": json, "svg": svg}
DEFAULT_FORMAT = "svg"


def render(report, fmt: str = DEFAULT_FORMAT, **opts):
    if fmt not in FORMATS:
        raise ValueError(f"unknown format: {fmt}")
    FORMATS[fmt].render(report, **opts)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import xml.etree.ElementTree as ET

from . import D
from ..bits import parse_cpu_set


def parse_report(path: pathlib.Path, data: D):
    data.vms = vms = D()
    for f in path.glob("etc/libvirt/qemu/*.xml"):
        xml = ET.parse(f).getroot()
        name = xml.find("./name").text
        vms[name] = vm = D(name=name)
        vm_cpu(vm, xml)
        vm_memory(vm, xml)
        vm_interfaces(vm, xml)


def vm_cpu(vm, xml):
    cpu = xml.find("./cpu")
    if cpu:
        vm.cpu_mode = cpu.get("mode")

    topo = xml.find("./cpu/topology")
    if topo:
        vm.topology = t = D()
        for attr in "sockets", "dies", "cores", "threads":
            t[attr] = int(topo.get(attr, "1"))

    for numa in xml.findall("./cpu/numa/cell"):
        numa_id = int(numa.get("id", "0"))
        vm.setdefault("numa", D())[numa_id] = D(
            id=numa_id,
            memory=int(numa.get("memory", "0")) * multiplier(numa.get("unit")),
            mem_access=numa.get("memAccess"),
            vcpus=parse_cpu_set(numa.get("cpus", str(numa_id))),
        )

    for vcpupin in xml.findall("./cputune/vcpupin"):
        vcpu = int(vcpupin.get("vcpu"))
        cpuset = parse_cpu_set(vcpupin.get("cpuset"))
        vm.setdefault("vcpu_pinning", D())[vcpu] = cpuset


def vm_memory(vm, xml):
    memory = xml.find("./memory")
    if memory:
        vm.memory = int(memory.text) * multiplier(memory.get("unit"))

    for node in xml.findall("./numatune/memnode"):
        numa_id = int(node.get("cellid", "0"))
        nodeset = parse_cpu_set(node.get("nodeset"))
        vm.setdefault("numa", D()).setdefault(numa_id, D()).host_numa = nodeset

    for page in xml.findall("./memoryBacking/hugepages/page"):
        node_set = parse_cpu_set(page.get("nodeset"))
        size = int(page.get("size")) * multiplier(page.get("unit"))
        for numa in vm.get("numa", D()).values():
            if node_set == numa.get("host_numa"):
                numa.hugepage_size = size


def vm_interfaces(vm, xml):
    for iface in xml.findall("./devices/interface"):
        iftype = iface.get("type")
        if iftype == "vhostuser":
            vm.setdefault("interfaces", []).append(
                D(
                    type=iftype,
                    queues=int(xpath(iface, "driver").get("queues", "1")),
                    socket=xpath(iface, "source").get("path"),
                )
            )
        elif iftype == "bridge":
            vm.setdefault("interfaces", []).append(
                D(
                    type="virtio-kernel",
                    bridge=xpath(iface, "source").get("bridge"),
                    net_dev=xpath(iface, "target").get("dev"),
                    ovs_port=xpath(iface, "virtualport/parameters").get("interfaceid"),
                )
            )
        elif iftype == "hostdev":
            src = iface.find("./source/address")
            domain = int(src.get("domain", "0x0"), 16)
            bus = int(src.get("bus", "0x0"), 16)
            slot = int(src.get("slot", "0x0"), 16)
            function = int(src.get("function", "0x0"), 16)
            dev = D(
                type="sriov",
                host_dev=f"{domain:04x}:{bus:02x}:{slot:02x}.{function:01x}",
            )
            for vlan in iface.findall("./vlan/tag"):
                dev.vlan = int(vlan.get("id"))
            vm.setdefault("interfaces", []).append(dev)

    for pci in xml.findall("./devices/hostdev[@type='pci']/source/address"):
        domain = int(pci.get("domain", "0x0"), 16)
        bus = int(pci.get("bus", "0x0"), 16)
        slot = int(pci.get("slot", "0x0"), 16)
        function = int(pci.get("function", "0x0"), 16)
        vm.setdefault("interfaces", []).append(
            D(
                type="sriov",
                host_dev=f"{domain:04x}:{bus:02x}:{slot:02x}.{function:01x}",
            )
        )


def xpath(node: ET.Element, path: str) -> ET.Element:
    n = node.find(path)
    if n is None:
        n = ET.Element(pathlib.Path(path).name)
    return n


def multiplier(unit):
    if isinstance(unit, str):
        if unit.startswith("Ki"):
            return 1024
        if unit.startswith("Mi"):
            return 1024 * 1024
        if unit.startswith("Gi"):
            return 1024 * 1024
    return 1



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import subprocess

from .dot import SOSGraph


def render(report, **opts):
    src = SOSGraph(report).source()
    subprocess.run(["dot", "-T", "svg"], input=src, text=True, check=True)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D
from ..bits import parse_cpu_set


INTERRUPT_RE = re.compile(r"^\s*(\w+):\s+([\s\d]+)\s+([A-Za-z].+)$", re.MULTILINE)


def parse_report(path: pathlib.Path, data: D):
    data.irqs = irqs = D()
    data.cpus = cpus = D()
    f = path / "proc/interrupts"
    if not f.is_file():
        return

    for match in INTERRUPT_RE.finditer(f.read_text()):
        irq, counters, desc = match.groups()
        irqs[irq] = D(
            irq=irq,
            desc=re.sub(r"\s+", " ", desc.strip()),
            counters=[int(c) for c in counters.split()],
        )
        try:
            irqs[irq].requested_affinity = parse_cpu_set(
                (path / f"proc/irq/{irq}/smp_affinity_list").read_text()
            )
            for cpu in irqs[irq].requested_affinity:
                c = cpus.setdefault(cpu, D(cpu=cpu))
                if "requested_irqs" in c:
                    c.requested_irqs += 1
                else:
                    c.requested_irqs = 1
            irqs[irq].effective_affinity = parse_cpu_set(
                (path / f"proc/irq/{irq}/effective_affinity_list").read_text()
            )
            for cpu in irqs[irq].effective_affinity:
                c = cpus.setdefault(cpu, D(cpu=cpu))
                if "effective_irqs" in c:
                    c.effective_irqs += 1
                else:
                    c.effective_irqs = 1
        except FileNotFoundError:
            pass



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D
from ..bits import parse_cpu_set


def parse_report(path: pathlib.Path, data: D):
    nodes = list(path.glob("sys/devices/system/node/node*"))
    for node in nodes:
        match = re.match(r"^node(\d+)$", node.name)
        if not match:
            continue

        numa_id = int(match.group(1))
        numa = data.setdefault("numa", D()).setdefault(numa_id, D(id=numa_id))

        cpus = parse_cpu_set((node / "cpulist").read_text())
        siblings = D()

        try:
            div = 1
            meminfo = (node / "meminfo").read_text()
        except FileNotFoundError:
            div = len(nodes)
            meminfo = (path / "proc/meminfo").read_text()

        match = re.search(r"MemTotal:\s+(\d+)\s*kB", meminfo)
        if match:
            numa.total_memory = int(int(match.group(1)) * 1024 / div)

        for huge in node.glob("hugepages/hugepages-*/nr_hugepages"):
            match = re.match(r"^hugepages-(\d+)kB$", huge.parent.name)
            if match:
                size = int(match.group(1)) * 1024
                numa.setdefault("hugepages", D())[size] = int(huge.read_text())

        for cpu in path.glob("sys/devices/system/cpu/cpu*/topology"):
            try:
                package_id = int((cpu / "physical_package_id").read_text())
                if package_id != numa_id:
                    continue
            except FileNotFoundError:
                cpu_id = int(re.match(r"cpu(\d+)", cpu.parent.name).group(1))
                if cpu_id not in cpus:
                    continue
            try:
                threads = parse_cpu_set((cpu / "thread_siblings_list").read_text())
            except FileNotFoundError:
                try:
                    threads = parse_cpu_set((cpu / "core_cpus_list").read_text())
                except FileNotFoundError:
                    # hyperthreading disabled
                    continue
            cpus.update(threads)
            for t in threads:
                siblings[t] = threads - {t}
        numa.cpus = cpus
        numa.thread_siblings = siblings



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import importlib
import pathlib
import pkgutil


class D(dict):

    def __getattr__(self, attr):
        return self[attr]

    def __setattr__(self, attr, value):
        return self.__setitem__(attr, value)


def parse_report(path: pathlib.Path) -> dict:
    data = D()
    for collector in discover_collectors():
        collector.parse_report(path, data)
    return data


def discover_collectors():
    for info in pkgutil.walk_packages(__path__, prefix=__name__ + "."):
        mod = importlib.import_module(info.name, __name__)
        if not (hasattr(mod, "parse_report") and callable(mod.parse_report)):
            continue
        yield mod



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry


import contextlib
import os
import re
import secrets

import graphviz

from ..bits import bit_list, human_readable
from ..collect import D


def render(report: D, **opts):
    print(SOSGraph(report).source())


class SOSGraph:

    def __init__(self, report: D):
        self.dot = graphviz.Graph(
            name="sosviz",
            node_attr={
                "fontsize": "11",
                "fontname": "monospace",
                "margin": "0.05",
                "shape": "rectangle",
            },
            edge_attr={
                "fontsize": "11",
                "fontname": "monospace",
                "margin": "0",
            },
            graph_attr={
                "fontsize": "11",
                "fontname": "monospace",
                "compound": "true",
                "rankdir": "LR",
            },
        )
        self.cur = self.dot
        self.stack = []
        self.links = set()
        self.clusters = set()
        self.report = report
        self.build()

    def source(self):
        return self.dot.source

    def safe_id(self, n):
        return re.sub(r"\W", "_", n)

    def edge(self, a, b, force=False, **kwargs):
        a = self.safe_id(a)
        b = self.safe_id(b)
        link = frozenset((a, b))
        if link in self.links and not force:
            # check for duplicate links
            return
        self.links.add(link)
        if "label" in kwargs:
            kwargs["label"] = format_label(kwargs["label"], max_width=30)
        self.dot.edge(a, b, **kwargs)

    @contextlib.contextmanager
    def subgraph(self, name=None, **kwargs):
        self.stack.append(self.cur)
        with self.cur.subgraph(name=name, graph_attr=kwargs) as cur:
            try:
                self.cur = cur
                yield cur
            finally:
                self.cur = self.stack.pop()

    def cluster(self, label, **kwargs):
        kwargs["label"] = format_label(label)
        if "style" not in kwargs:
            kwargs["style"] = "dotted"
        kwargs["cluster"] = "true"
        if isinstance(label, list):
            label = label[0]
        name = self.safe_id(label)
        if name in self.clusters:
            name += secrets.token_hex(8)
        self.clusters.add(name)
        return self.subgraph(name=name, **kwargs)

    def group(self, **kwargs):
        kwargs["cluster"] = "false"
        return self.subgraph(name=None, **kwargs)

    def node(self, name: str, label: str, **kwargs):
        name = self.safe_id(name)
        if kwargs.get("shape", "rectangle") != "rectangle":
            kwargs.setdefault("margin", "0")
        if "tooltip" in kwargs:
            kwargs["tooltip"] = format_label(kwargs["tooltip"])
        self.cur.node(name, format_label(label, max_width=30), **kwargs)

    def build(self):
        r = self.report
        label = [
            f"<b>{r.hostname} / <i>{r.hardware.system}</i></b>",
            f"<i>linux {r.software.kernel}</i>",
        ]
        for sw in "os_release", "rhosp_release":
            if sw in r.software:
                label.append(r.software.get(sw))

        if "ovs" in r:
            ovs = f"OVS {r.ovs.config.ovs_version}"
            if r.ovs.config.get("dpdk_initialized"):
                ovs += f" {r.ovs.config.dpdk_version}"
            label.append(ovs)

        with self.cluster(label):
            # vms
            for vm in r.get("vms", {}).values():
                with self.cluster(f"VM {vm.name}", style="solid"):
                    for i, numa in vm.get("numa", {}).items():
                        with self.cluster(f"numa {i}", style="dotted"):
                            with self.group(rank="source"):
                                self.vm_numa(vm, numa)
                    for iface in vm.get("interfaces", []):
                        self.vm_iface(iface)

            # linux interfaces
            with self.cluster("networking"):
                # openvswitch
                for br in r.get("ovs", D()).get("bridges", D()).values():
                    self.ovs_bridge(br)
                for port in r.get("ovs", D()).get("ports", D()).values():
                    if port.type in ("vxlan", "geneve", "internal", ""):
                        continue
                    if port.type == "patch":
                        self.edge(
                            self.ovs_br_node_id(port.bridge),
                            self.ovs_br_node_id(r.ovs.ports[port.options.peer].bridge),
                            style="dashed",
                            color="forestgreen",
                        )
                        continue
                    self.ovs_port(port)

                # linux networking
                for iface in r.get("interfaces", D()).values():
                    self.phy_iface(iface, netns="")

                for netns, ifaces in r.netns.items():
                    with self.cluster(f"netns {netns}", color="salmon", style="dashed"):
                        for iface in ifaces.values():
                            self.phy_iface(iface, netns=netns)

            # physical CPU/memory
            for numa in r.get("numa", D()).values():
                with self.cluster(f"phy numa {numa.id}"):
                    self.phy_numa(numa)

    def vm_cpu_node_id(self, vm, numa):
        return f"{vm.name}_cpu_{numa.id}"

    def vm_memory_node_id(self, vm, numa):
        return f"{vm.name}_memory_{numa.id}"

    def vm_numa(self, vm: D, numa: D):
        labels = [f"<b>vCPUs {bit_list(numa.vcpus)}</b>"]
        host_cpus = set()
        for vcpu in numa.vcpus:
            host_cpus.update(vm.vcpu_pinning[vcpu])
        labels.append(f"host CPUs {bit_list(host_cpus)}")
        cpu_numas = set()
        for n in self.report.numa.values():
            if host_cpus & n.cpus:
                cpu_numas.add(n.id)
        labels.append(f'<font color="blue">host NUMA {bit_list(cpu_numas)}</font>')
        self.node(self.vm_cpu_node_id(vm, numa), labels, color="blue")

        labels = [f"<b>memory {human_readable(numa.memory, 1024)}</b>"]
        for h in numa.get("host_numa", []):
            labels.append(f'<font color="red">host NUMA {h}</font>')
        self.node(self.vm_memory_node_id(vm, numa), labels, color="red")

    def vm_iface_node_id(self, name):
        return f"vm_iface_{name}"

    def vm_iface(self, iface: D) -> str:
        labels = [f"<b>{iface.type}</b>"]
        name = None
        peer = None
        color = "darkorange"
        if "socket" in iface:
            color = "forestgreen"
            name = os.path.basename(iface.socket)
            peer = self.ovs_port_node_id(name)
        elif "host_dev" in iface:
            color = "darkorange"
            name = iface.host_dev
            peer = pci_node_id(iface.host_dev)
        elif "net_dev" in iface:
            color = "hotpink"
            name = iface.net_dev
            peer = self.iface_node_id(iface.net_dev, "")
        if name:
            labels.append(name)
        if "vlan" in iface:
            labels.append(f"VLAN {iface.vlan}")
        if "bridge" in iface:
            labels.append(f"<i>bridge {iface.bridge}</i>")
        if "queues" in iface:
            labels.append(f"<i>queues {iface.queues}</i>")
        self.node(self.vm_iface_node_id(name), labels, color=color)
        if peer:
            self.edge(self.vm_iface_node_id(name), peer, style="dashed", color=color)

    def iface_node_id(self, name, netns):
        return f"net_{netns}_{name}"

    def phy_iface(self, iface: D, netns: str = ""):
        if netns == "" and iface.name in self.report.ovs.bridges:
            return
        if iface.name in "ovs-system" or re.match(r"(genev|vxlan)_sys_\d+", iface.name):
            return

        labels = [f"<b>{iface.name}</b>"]

        kind = iface.get("kind")
        if kind == "bond" and "bond_mode" in iface:
            kind += f" {iface.bond_mode}"
        if kind == "vlan" and "vlan" in iface:
            kind += f" {iface.vlan}"
        if kind == "tun" and "tun_type" in iface:
            kind += f" {iface.tun_type}"
        if kind:
            labels.append(kind)
        if "device" in iface:
            labels.append(iface.device)
            self.edge(
                self.iface_node_id(iface.name, netns),
                pci_node_id(iface.device),
                style="dashed",
                color="darkorange",
            )

        for ip in iface.get("ip", []):
            labels.append(f'<font color="purple">{ip}</font>')

        tooltips = [iface["flags"]]
        if "mac" in iface:
            tooltips.append(f"mac {iface.mac}")
        if "mtu" in iface:
            tooltips.append(f"mtu {iface.mtu}")

        color = "salmon" if netns else "hotpink"
        self.node(
            self.iface_node_id(iface.name, netns),
            labels,
            color=color,
            tooltip=tooltips,
        )
        if "link" in iface:
            if "link_netns" in iface:
                link_ns = iface.link_netns
                if link_ns == "0":
                    link_ns = ""
            else:
                link_ns = netns
            link = self.find_iface(iface.link, link_ns)
            self.edge(
                self.iface_node_id(iface.name, netns),
                self.iface_node_id(link.name, link_ns),
                style="dashed",
                color=color,
            )
        if "master" in iface and iface.master != "ovs-system":
            self.edge(
                self.iface_node_id(iface.master, netns),
                self.iface_node_id(iface.name, netns),
                style="solid",
                color=color,
            )
        if iface.name in self.report.ovs.ports:
            port = self.report.ovs.ports[iface.name]
            self.edge(
                self.ovs_br_node_id(port.bridge),
                self.iface_node_id(iface.name, netns),
                style="solid",
                color="forestgreen",
            )

    def find_iface(self, name, netns):
        if netns:
            ifaces = self.report.netns[netns]
        else:
            ifaces = self.report.interfaces
        match = re.match(r"if(\d+)", name)
        if not match:
            return ifaces[name]
        for iface in ifaces.values():
            if iface.index == match.group(1):
                return iface
        raise KeyError(name)

    def ovs_br_node_id(self, name):
        return f"ovs_br_{name}"

    def ovs_bridge(self, br: D):
        labels = [
            f"<b>{br.name}</b>",
            f"OVS {br.datapath}",
            f"rules {br.of_rules}",
            f"ports {br.ports}",
        ]
        iface = self.report.interfaces.get(br.name, None)
        if iface is not None:
            for ip in self.report.interfaces[br.name].get("ip", []):
                labels.append(f'<font color="purple">{ip}</font>')
            if "link" in iface:
                link_ns = ""
                if "link_netns" in iface:
                    link_ns = iface.link_netns
                    if link_ns == "0":
                        link_ns = ""
                link = self.find_iface(iface.link, link_ns)
                self.edge(
                    self.ovs_br_node_id(br.name),
                    self.iface_node_id(link.name, link_ns),
                    style="dashed",
                    color="forestgreen",
                )
            if "master" in iface:
                self.edge(
                    self.iface_node_id(iface.master, ""),
                    self.ovs_br_node_id(br.name),
                    style="solid",
                    color="forestgreen",
                )
        self.node(
            self.ovs_br_node_id(br.name), labels, color="forestgreen", shape="diamond"
        )
        vxlan_ports = 0
        geneve_ports = 0
        for port in self.report.ovs.ports.values():
            if port.bridge != br.name:
                continue
            if port.type == "vxlan":
                vxlan_ports += 1
            if port.type == "geneve":
                geneve_ports += 1
        if vxlan_ports > 0:
            vxlan_stub = self.ovs_br_node_id(br.name) + "_vxlans"
            self.node(
                vxlan_stub,
                f"{vxlan_ports} VXLAN ports",
                shape="ellipse",
                color="forestgreen",
                style="dashed",
            )
            self.edge(
                vxlan_stub,
                self.ovs_br_node_id(br.name),
                color="forestgreen",
                style="solid",
            )
        if geneve_ports > 0:
            geneve_stub = self.ovs_br_node_id(br.name) + "_geneves"
            self.node(
                geneve_stub,
                f"{geneve_ports} GENEVE ports",
                shape="ellipse",
                color="forestgreen",
                style="dashed",
            )
            self.edge(
                geneve_stub,
                self.ovs_br_node_id(br.name),
                color="forestgreen",
                style="solid",
            )

    def ovs_port_node_id(self, name):
        return f"ovs_port_{name}"

    def ovs_dpdk_labels(self, port: D):
        if "dpdk_devargs" in port.options:
            numa_id = "N/A"
            for numa in self.report.numa.values():
                if port.options.dpdk_devargs in numa.pci_nics:
                    numa_id = numa.id
                    break
            yield f"{port.options.dpdk_devargs} NUMA {numa_id}"

        disabled = set()
        rxqs = []
        for pmd in self.report.ovs.pmds.values():
            for rxq in pmd.rxqs:
                if rxq.port == port.name:
                    if rxq.enabled:
                        rxqs.append((rxq.rxq, pmd.core, pmd.numa))
                    else:
                        disabled.add(rxq.rxq)

        for rxq, core, numa in sorted(rxqs):
            yield f"rxq {rxq} cpu {core} NUMA {numa}"
        if disabled:
            s = "s" if len(disabled) > 1 else ""
            yield f'<font color="gray">rxq{s} {bit_list(disabled)} disabled</font>'

    def ovs_port(self, port: D):
        if port.name in self.report.interfaces:
            self.edge(
                self.iface_node_id(port.name, ""),
                self.ovs_br_node_id(port.bridge),
                color="forestgreen",
            )
            return

        labels = [
            f"<b>{port.name}</b>",
            f"OVS {port.type}",
        ]
        if port.type in ("dpdk", "dpdkvhostuserclient"):
            labels.extend(self.ovs_dpdk_labels(port))
            if "dpdk_devargs" in port.options:
                self.edge(
                    self.ovs_port_node_id(port.name),
                    pci_node_id(port.options.dpdk_devargs),
                    style="dashed",
                    color="darkorange",
                )

        if "tag" in port:
            labels.append(f'<font color="forestgreen">VLAN {port.tag}</font>')
        iface = self.report.interfaces.get(port.name, None)
        if iface is not None:
            for ip in self.report.interfaces[port.name].get("ip", []):
                labels.append(f'<font color="purple">{ip}</font>')
            if "link" in iface:
                link_ns = ""
                if "link_netns" in iface:
                    link_ns = iface.link_netns
                    if link_ns == "0":
                        link_ns = ""
                link = self.find_iface(iface.link, link_ns)
                self.edge(
                    self.ovs_port_node_id(port.name),
                    self.iface_node_id(link.name, link_ns),
                    style="dashed",
                    color="forestgreen",
                )
            if "master" in iface:
                self.edge(
                    self.iface_node_id(iface.master, ""),
                    self.ovs_port_node_id(port.name),
                    style="solid",
                    color="forestgreen",
                )

        self.node(
            self.ovs_port_node_id(port.name),
            labels,
            color="forestgreen",
        )
        self.edge(
            self.ovs_br_node_id(port.bridge),
            self.ovs_port_node_id(port.name),
            color="forestgreen",
        )

        if port.type == "bond":
            for member in port.members.values():
                if member.name in self.report.interfaces:
                    self.edge(
                        self.ovs_port_node_id(port.name),
                        self.iface_node_id(member.name, ""),
                        style="solid",
                        color="forestgreen",
                    )
                    continue

                labels = [
                    f"<b>{member.name}</b>",
                    f"OVS {member.type}",
                ]
                if member.type == "dpdk":
                    labels.extend(self.ovs_dpdk_labels(member))
                    self.edge(
                        self.ovs_port_node_id(member.name),
                        pci_node_id(member.options.dpdk_devargs),
                        style="dashed",
                        color="darkorange",
                    )
                self.node(
                    self.ovs_port_node_id(member.name),
                    labels,
                    color="forestgreen",
                )
                self.edge(
                    self.ovs_port_node_id(port.name),
                    self.ovs_port_node_id(member.name),
                    style="solid",
                    color="forestgreen",
                )

    def irq_counters(self, cpu: int) -> tuple[int, int]:
        counter = 0
        bound = 0
        for irq in self.report.irqs.values():
            if not irq.irq.isdigit():
                continue
            if cpu in irq.get("effective_affinity", []):
                bound += 1
            counter += irq.counters[cpu]
        return counter, bound

    def irq_counters_tooltip(self, cpus: set[int]) -> str:
        tooltip = []
        for c in cpus:
            counter, bound = self.irq_counters(c)
            if counter < 42:  # XXX: how about 1337 maybe?
                continue
            counter = human_readable(counter)
            tooltip.append(f"CPU {c} bound_irqs={bound} interrupts={counter}")
        return format_label(tooltip)

    def phy_numa(self, numa: D):
        model = "<b>Unknown Processor Model</b>"
        for i, proc in enumerate(self.report.hardware.processor):
            if i == numa.id:
                model = f"<b>{proc.model}</b>"
        with self.cluster(model, style="dotted", color="blue"):
            housekeeping_cpus = set(numa.cpus)

            ovs_pmds = {}
            for pmd in self.report.ovs.pmds.values():
                if pmd.numa != numa.id:
                    continue
                p = ovs_pmds.setdefault(
                    pmd.core,
                    D(cpu=pmd.core, isolated=pmd.isolated, rxqs=0, usage=0),
                )
                for rxq in pmd.rxqs:
                    if rxq.enabled:
                        p.rxqs += 1
                        p.usage += rxq.usage

            if ovs_pmds:
                labels = ["<b>OVS DPDK</b>"]
                for pmd in ovs_pmds.values():
                    irqs, _ = self.irq_counters(pmd.cpu)
                    labels.append(
                        f"CPU {pmd.cpu} rxqs={pmd.rxqs} usage={pmd.usage}% irqs={human_readable(irqs)}"
                    )

                self.node(
                    f"phy_cpus_ovs_{numa.id}",
                    labels,
                    tooltip=self.irq_counters_tooltip(ovs_pmds.keys()),
                    color="blue",
                )

            housekeeping_cpus -= ovs_pmds.keys()

            for vm in self.report.get("vms", {}).values():
                for vnuma in vm.numa.values():
                    if numa.id not in vnuma.host_numa:
                        continue
                    host_cpus = set()
                    for vcpu in vnuma.vcpus:
                        host_cpus.update(vm.vcpu_pinning[vcpu])
                    self.node(
                        f"phy_cpus_{vm.name}_{numa.id}",
                        [f"<b>VM {vm.name}</b>", f"CPUs {bit_list(host_cpus)}"],
                        tooltip=self.irq_counters_tooltip(host_cpus),
                        color="blue",
                    )
                    housekeeping_cpus -= host_cpus

            self.node(
                f"phy_cpus_housekeeping_{numa.id}",
                ["<b>Housekeeping</b>", f"CPUs {bit_list(housekeeping_cpus)}"],
                tooltip=self.irq_counters_tooltip(housekeeping_cpus),
                color="blue",
            )

        labels = [f"<b>memory {human_readable(numa.total_memory, 1024)}</b>"]
        for size, num in numa.get("hugepages", {}).items():
            if not num:
                continue
            labels.append(f"{human_readable(size, 1024)} hugepages: {num}")

        self.node(f"memory_{numa.id}", labels, color="red")
        self.phy_pci_nics(numa)

    def phy_pci_nics(self, numa: D):
        pci_bridges = {}
        for nic in numa.get("pci_nics", {}).values():
            pci_bridges.setdefault(nic.pci_bridge, []).append(nic)

        for bridge, nics in pci_bridges.items():
            if len(nics) == 1:
                labels = [f"<b>{nics[0].pci_id}</b>"]
                if "kernel_driver" in nics[0]:
                    labels.append(nics[0].kernel_driver)
                self.node(
                    pci_node_id(nics[0].pci_id),
                    labels,
                    tooltip=nics[0].device,
                    color="darkorange",
                )
            else:
                with self.cluster(
                    f"PCI bridge {bridge}", style="dashed", color="darkorange"
                ):
                    for nic in nics:
                        labels = [f"<b>{nic.pci_id}</b>"]
                        if "kernel_driver" in nic:
                            labels.append(nic.kernel_driver)
                        self.node(
                            pci_node_id(nic.pci_id),
                            labels,
                            tooltip=nic.device,
                            color="darkorange",
                        )


def pci_node_id(pci_id):
    return f"pci_{pci_id}"


def wrap_text(text: str, margin: int) -> list[str]:
    lines = []
    more = True

    while more:
        if len(text) <= margin:
            # whole text fits in a single line
            line = text
            more = False
        else:
            # find split point, preferably before margin
            split = -1
            width = 0
            markup = 0
            for w, t in enumerate(text):
                if width >= margin and split != -1:
                    break
                if t in " \t,>":
                    split = w
                if t == "<":
                    markup += 1
                elif t == ">":
                    markup -= 1
                if markup == 0:
                    width += 1
            if split == -1:
                # no space found to split, print a long line
                line = text
                more = False
            else:
                line = text[: split + 1]
                text = text[split + 1 :]
                # find start of next word
                while text and text[0] in " \t":
                    text = text[1:]
                if not text:
                    # only trailing whitespace, we're done
                    more = False
        lines.append(line)

    return lines


def format_label(lines: list[str], max_width: int = 0) -> str:
    if isinstance(lines, str):
        lines = [lines]
    if max_width:
        out = []
        for line in lines:
            out += wrap_text(line, max_width)
        lines = out
    if any(line.startswith("<") for line in lines):
        return "<" + "<br/>".join(lines) + ">"
    return "\\n".join(lines)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D


def parse_report(path: pathlib.Path, data: D):
    data.hardware = hw = D(system="Unknown Hardware", processor=[], memory=[])
    f = path / "sos_commands/hardware/dmidecode"
    if not f.is_file():
        return
    dmi = f.read_text()
    blocks = re.split(
        r"^Handle 0x[a-fA-F0-9]+, DMI type \d+, \d+ bytes$", dmi, flags=re.MULTILINE
    )
    for block in blocks:
        if "\n" not in block.strip():
            continue
        name, block = block.strip().split("\n", 1)
        if name == "System Information":
            b = split_block(block)
            hw.system = f"{b.Manufacturer} {b['Product Name']}"

        elif name == "Memory Device":
            b = split_block(block)
            if "Speed" not in b:
                continue
            hw.memory.append(
                D(
                    type=f"{b.Type} {b['Type Detail']}",
                    size=b.Size,
                    slot=b.Locator,
                    speed=b.Speed,
                )
            )

        elif name == "Processor Information":
            b = split_block(block)
            if "Version" not in b:
                continue
            hw.processor.append(
                D(
                    slot=b["Socket Designation"],
                    model=b.Version,
                    cores=int(b["Core Enabled"]),
                    threads=int(b["Thread Count"]),
                )
            )


def split_block(block: str) -> D:
    fields = D()
    for match in re.finditer(r"^\t([\w\s-]+):\s(.*)$", block, flags=re.MULTILINE):
        fields[match.group(1).strip()] = match.group(2).strip()
    return fields



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D


def parse_report(path: pathlib.Path, data: dict):
    bridges = pci_bridges(path)
    for node in path.glob("sys/devices/system/node/node*"):
        match = re.match(r"node(\d+)", node.name)
        if not match:
            continue

        numa_id = int(match.group(1))
        numa = data.setdefault("numa", D()).setdefault(numa_id, D(id=numa_id))
        nics = numa.setdefault("pci_nics", D())

        lspci = (path / "sos_commands/pci/lspci_-nnvv").read_text()
        for block in lspci.split("\n\n"):
            if not re.search(rf"^\tNUMA node: {numa_id}$", block, flags=re.MULTILINE):
                continue

            match = re.search(
                r"^([a-f0-9:\.]+) Ethernet controller \[0200\]: (.*)$",
                block,
                flags=re.MULTILINE,
            )
            if not match:
                continue

            pci_id = match.group(1)
            if len(pci_id) != len("0000:00:00.0"):
                pci_id = "0000:" + pci_id

            nic = nics.setdefault(pci_id, D())
            nic.pci_id = pci_id
            nic.device = match.group(2)
            k = re.search(r"^\tKernel driver in use: (.+)$", block, flags=re.MULTILINE)
            if k is not None:
                nic.kernel_driver = k.group(1)
            for net in path.glob(f"sys/devices/pci*/*/{pci_id}/net/*"):
                nic.netdev = net.name
            nic.pci_bridge = bridges.get(nic.pci_id[: len("0000:00")])


PCI_TREE_L1 = r"^[\s-][\+\\]-\[(?P<l1>[a-f\d]{4}:[a-f\d]{2})\]-"
PCI_TREE_L2 = r"^[\|\s]*[\+\\]-(?P<l2>[a-f\d]{2}\.[a-f\d])"
PCI_TREE_L3 = r"^[\|\s]*-\[(?P<l3start>[a-f\d]{2})(?:-(?P<l3end>[a-f\d]{2}))?\]"


def pci_bridges(path: pathlib.Path) -> dict:
    bridges = {}
    l1 = None
    l2 = None

    for line in (path / "sos_commands/pci/lspci_-tv").read_text().splitlines():
        match = re.search(PCI_TREE_L1, line)
        if match:
            l1 = match.group("l1")
            line = line[match.end() :]
        match = re.search(PCI_TREE_L2, line)
        if match:
            l2 = match.group("l2")
            line = line[match.end() :]
        match = re.search(PCI_TREE_L3, line)
        if match and l1 and l2:
            l3start = match.group("l3start")
            l3end = match.group("l3end") or l3start
            start = int(l3start, 16)
            end = int(l3end, 16)
            bridge_addr = f"{l1}:{l2}"
            domain, _, _ = l1.partition(":")
            for prefix in range(start, end + 1):
                bridges[f"{domain}:{prefix:02x}"] = bridge_addr

    return bridges



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D


IFACE_RE = re.compile(
    r"""
    ^(?P<index>\d+):\s+(?P<name>[^@:]+?)(?:@(?P<link>[^:]+))?:\s+
        <(?P<flags>[\w,-]+)>\s+
        mtu\s+(?P<mtu>\d+)
        (?:.*?\smaster\s+(?P<master>\S+)\s)?.*\n
    ^\s+link/ether\s+(?P<mac>[a-f\d:]+)\sbrd\s[a-f\d:]+
        (?:\s+link-netns(?:id)?\s+(?P<link_netns>\S+))?\s.*\n
    (?:^\s+
        (?P<kind>
            vlan
            |
            bridge
            |
            bond
            |
            veth
            |
            vxlan
            |
            tun
            |
            bond_slave
            |
            bridge_slave
        )\s+
        (?:
            protocol\s+\S+\s+id\s+(?P<vlan>\d+)
            |
            type\s+(?P<tun_type>\S+)
            |
            mode\s+(?P<bond_mode>\S+)
            |
            state\s+(?P<slave_state>\S+)
        )?.*
    \n)?
    """,
    re.VERBOSE | re.MULTILINE,
)
ADDR_RE = re.compile(
    r"inet6? (?P<addr>(?:\d+(?:\.\d+){3}|[a-f\d:]+)/\d+) .*scope global",
)
IFACE_BLOCK_RE = re.compile(r"^\d+:\s.*\n(\s{4}.+\n)+", re.VERBOSE | re.MULTILINE)


def parse_report(path: pathlib.Path, data: D):
    data.interfaces = parse_interfaces(
        path / "sos_commands/networking/ip_-d_address", path
    )
    data.netns = D()
    for netns in path.glob("sos_commands/networking/namespaces/*/*_ip_-d_address_show"):
        data.netns[netns.parent.name] = parse_interfaces(netns, path)


def parse_interfaces(ip_addr: pathlib.Path, path: pathlib.Path) -> D:
    ifaces = D()
    if not ip_addr.is_file():
        return ifaces
    for block in IFACE_BLOCK_RE.finditer(ip_addr.read_text()):
        match = IFACE_RE.search(block.group())
        if not match:
            continue
        d = D({k: v for (k, v) in match.groupdict().items() if v is not None})
        for dev in path.glob(f"sys/class/net/{d.name}/device"):
            d.device = dev.resolve().name
        ifaces[d.name] = d
        for m in ADDR_RE.finditer(block.group()):
            d.setdefault("ip", []).append(m.group("addr"))
    return ifaces



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import re
import typing


HEX_RE = re.compile(r"0x[0-9a-fA-F]+")
RANGE_RE = re.compile(r"\d+-\d+")
INT_RE = re.compile(r"\d+")


def parse_cpu_set(arg: str) -> typing.Set[int]:
    bit_ids = set()

    for item in arg.strip().split(","):
        if not item:
            continue
        if HEX_RE.match(item):
            item = int(item, 16)
            bit = 0
            while item != 0:
                if item & 1:
                    bit_ids.add(bit)
                bit += 1
                item >>= 1
        elif RANGE_RE.match(item):
            start, end = item.split("-")
            bit_ids.update(range(int(start, 10), int(end, 10) + 1))
        elif INT_RE.match(item):
            bit_ids.add(int(item, 10))
        else:
            raise ValueError(f"invalid cpu set: {item}")
    return bit_ids


def hex_mask(bit_ids: typing.Set[int]) -> str:
    mask = 0
    for bit in bit_ids:
        mask |= 1 << bit
    return hex(mask)


def bit_mask(bit_ids: typing.Set[int]) -> str:
    mask = 0
    for bit in bit_ids:
        mask |= 1 << bit
    return f"0b{mask:_b}"


def bit_list(bit_ids: typing.Set[int]) -> str:
    groups = []
    bit_ids = sorted(bit_ids)
    i = 0
    while i < len(bit_ids):
        low = bit_ids[i]
        while i < len(bit_ids) - 1 and bit_ids[i] + 1 == bit_ids[i + 1]:
            i += 1
        high = bit_ids[i]
        if low == high:
            groups.append(str(low))
        elif low + 1 == high:
            groups.append(f"{low},{high}")
        else:
            groups.append(f"{low}-{high}")
        i += 1
    return ",".join(groups)


def human_readable(value: float, order: int = 1000) -> str:
    units = ("K", "M", "G")
    i = 0
    unit = ""
    while value >= order and i < len(units):
        unit = units[i]
        value /= order
        i += 1
    if unit == "":
        return str(value)
    if value < 100 and value % 1 > 0:
        return f"{value:.1f}{unit}"
    return f"{value:.0f}{unit}"



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import pathlib
import re

from . import D
from ..bits import parse_cpu_set


VARIABLE_RE = re.compile(r"^(\w+)\s*=\s*(.+)$", re.MULTILINE)


def parse_report(path: pathlib.Path, data: D):
    cmdline = (path / "proc/cmdline").read_text()
    tuning = D()
    for prop in "isolcpus", "nohz_full", "rcu_nocbs":
        match = re.search(rf"{prop}=([0-9,-]+)", cmdline)
        if match:
            tuning[prop] = parse_cpu_set(match.group(1))

    tuned_profile = path / "etc/tuned/active_profile"
    if tuned_profile.is_file():
        profile = tuned_profile.read_text().strip()
        tuning.tuned = D(profile=profile, variables=D())
        variables = path / f"etc/tuned/{profile}-variables.conf"
        if variables.is_file():
            for match in VARIABLE_RE.finditer(variables.read_text()):
                tuning.tuned.variables[match.group(1)] = match.group(2)

    irqbalance = path / "etc/sysconfig/irqbalance"
    if irqbalance.is_file():
        cpus = set()
        for match in VARIABLE_RE.finditer(irqbalance.read_text()):
            name, value = match.groups()
            if name == "IRQBALANCE_BANNED_CPULIST":
                cpus.update(parse_cpu_set(value))
            elif name == "IRQBALANCE_BANNED_CPUS":
                value = "0x" + value.replace(",", "").strip()
                cpus.update(parse_cpu_set(value))
        tuning.irq_banned_cpus = cpus

    data.tuning = tuning



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import json


def render(report, **opts):
    print(json.dumps(report, default=cast_json))


def cast_json(obj):
    if isinstance(obj, set):
        return list(obj)
    return obj



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

import os
import pprint


def render(report, **opts):
    try:
        width, _ = os.get_terminal_size()
    except OSError:
        width = 100
    pprint.pprint(report, compact=True, width=width)



# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Robin Jarry

"""
Collect information from an sos report folder and export it in other formats
on standard output.
"""

import argparse
from importlib import metadata
import pathlib
import sys

from . import collect, output


def main():
    parser = argparse.ArgumentParser(description=__doc__, prog="sosviz")
    parser.add_argument(
        "path",
        metavar="PATH",
        type=pathlib.Path,
        help="""
        Path to an uncompressed sos report folder.
        """,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {metadata.version('sosviz')}",
        help="""
        Show version and exit.
        """,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="""
        Show debug info.
        """,
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=output.FORMATS.keys(),
        default=output.DEFAULT_FORMAT,
        help="""
        Output format (default: %(default)s).
        """,
    )
    args = parser.parse_args()
    try:
        if not args.path.is_dir():
            raise ValueError(f"'{args.path}': No such directory")
        report = collect.parse_report(args.path)
        output.render(report, args.format)
    except BrokenPipeError:
        pass
    except Exception as e:
        if args.debug or isinstance(e, NotImplementedError):
            raise
        print(f"error: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()
